package tomracer;


import java.io.FileOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import org.jdom.Element;
import org.jdom.output.XMLOutputter;

import com.sun.corba.se.impl.ior.NewObjectKeyTemplateBase;

import simplerace.BasicTrack;
import simplerace.CompetitionScore;
import simplerace.Constants;
import simplerace.Controller;
import simplerace.Evaluator;
import simplerace.HeuristicController;
import simplerace.KeyboardControllerOne;
import simplerace.SensorModel;
import simplerace.Utils;
import simplerace.Vector2d;
import tomracer.genetics.Genome;
import tomracer.neural.CTRNNNeuron;
import tomracer.neural.CellConnection;
import tomracer.neural.InputNeuron;
import tomracer.neural.NeuralNetwork;
import tomracer.neural.Neuron;
import tomracer.neural.OutputNeuron;
import tomracer.neural.SigmoidNeuron;
import tomracer.neural.TanHNeuron;
import tomracer.util.Pair;
import wox.serial.ObjectWriter;
import wox.serial.SimpleWriter;
import wox.serial.TestObject;

public class CompositeSimpleRacer implements INeuralGeneticRacer {
	

	protected static boolean showInputsAndOutputs=false;

	NeuralNetwork speedController, directionController, targetChooser;
	
	static double[] evolvedGenes = {-0.95,-1.48,1.20,0.75,-0.91,-1.54,0.42,1.74,1.28,-1.30,-0.02,1.23,1.42,0.47,-0.24,1.04,-0.10,0.21,1.95,1.39,1.79,1.16,-1.80,1.37,-0.85,1.36,1.53,-1.27,0.62,0.33,-0.62,-1.69,-0.48,-0.57,-0.63,-0.00,-0.26,-0.75,-0.70,0.91,0.53,0.88,-0.96,-0.95,0.47,1.43,-0.04,-1.73,-0.15,1.52,-1.10,-1.28,0.81,1.44,-1.70,-0.41,0.91,-0.44,-0.97,0.93,0.88,-1.92,-1.55,-0.58,-0.15,-0.98,1.29,1.26,0.38,-0.73,-1.52,0.58,0.99,-0.97,-0.03,-0.13,1.90,0.38,0.90,1.44,-1.49,-1.38,-1.39,-1.12,1.53,-1.74,1.69,-0.32,1.23,1.99,1.67,1.95,-1.67,0.62,-1.98,1.20,1.84,-1.94,1.89,-0.49,0.50,-1.57,0.14,-0.23,-0.17,-0.12,-1.47,-0.86,0.43,-0.15,-0.24,-0.28,-1.75,-1.02,0.69,0.32,1.82,0.40,0.14,-0.26,1.40,0.45,-0.58,1.93,-1.99,0.37,-0.84,1.56,0.19,-0.54,-0.50,-0.62,1.11,1.54,1.72,-1.26,-1.78,0.52,-1.20,1.82,-1.82,0.35,1.06,1.63,0.69,1.98,-0.91,0.70,-0.23,-1.93,0.08,1.65,-1.64,-1.21,0.70,0.76,1.98,-0.45,1.11,1.87,1.68,-1.23,-0.10,1.15,-1.68,-1.25,1.36,0.30,0.60,0.47,1.40,-0.64,0.25,-1.73,0.57,-1.32,-1.04,1.90,1.68,-1.76,-1.33,-0.59,0.90,-0.03,0.54,-1.94,0.26,-0.51,-1.53,-1.56,-0.39,-1.25,-0.05,-1.33,1.68,0.84,-0.97,-0.64,-0.54,-0.85,-0.75,-1.97,0.82,0.78,0.14,0.08,0.65,0.44,-1.33,1.61,-0.89,0.46,1.00,-1.05,1.01,-0.23,-1.58,0.99,-0.90,0.42,1.02,1.58,1.62,0.97,-0.80};

	static Genome sampleChromosome = CompositeSimpleRacer.getSampleChromosome(true);
	static {
		sampleChromosome.setGenes(evolvedGenes);
	}

	public CompositeSimpleRacer() {
		this(sampleChromosome,true,true);
	}
	
	int genePos=0;
	public CompositeSimpleRacer(Genome chromosome, boolean isRecurrent, boolean useCTRNN) {
		createNetworks(chromosome, isRecurrent,useCTRNN,false);
	}

	public int createNetworks(Genome chromosome, boolean isRecurrent, boolean useCTRNN, boolean simulation) {
		Pair<NeuralNetwork, Integer> networkPair = createNetwork(chromosome, genePos, isRecurrent, useCTRNN, numInputs_SC(), numHidden_SC(), numOutputs_SC(), simulation);
		speedController = networkPair.first();		
		networkPair = createNetwork(chromosome, networkPair.second(), isRecurrent, useCTRNN, numInputs_DC(), numHidden_DC(), numOutputs_DC(),simulation);
		directionController = networkPair.first();		
		networkPair = createNetwork(chromosome, networkPair.second(), isRecurrent, useCTRNN, numInputs_TC(), numHidden_TC(), numOutputs_TC(),simulation);
		targetChooser = networkPair.first();
		return networkPair.second();
	}

	public static int numInputs_SC() {
		return 7;
	}

	public static int numHidden_SC() {
		return 5;
	}

	public static int numOutputs_SC() {
		return 1;
	}
	
	public static boolean isReccurent_SC() {
		return true;
	}

	public static int numInputs_DC() {
		return 7;
	}

	public static int numHidden_DC() {
		return 5;
	}

	public static int numOutputs_DC() {
		return 1;
	}
	
	public static boolean isReccurent_DC() {
		return true;
	}

	public static int numInputs_TC() {
		return 7;
	}

	public static int numHidden_TC() {
		return 5;
	}

	public static int numOutputs_TC() {
		return 1;
	}
	
	public static boolean isReccurent_TC() {
		return true;
	}

	
	private static Pair<NeuralNetwork, Integer> createNetwork(Genome chromosome, int genePos, boolean isRecurrent, boolean useCTRNN, int numInputs, int numHidden, int numOutputs, boolean simulation) {
		NeuralNetwork net = simulation ? null : new NeuralNetwork();
			
		List<Neuron> inputNeurons = simulation ? null : new ArrayList<Neuron>();
		
		if (!simulation)
		for (int i=0; i < numInputs; i++) {
			InputNeuron inputNeuron = new InputNeuron();
			inputNeurons.add(inputNeuron);
			net.addInputNeuron(inputNeuron);

		}

		List<Neuron> hiddenNeurons = simulation ? null: new ArrayList<Neuron>();
		for (int i = 0; i < numHidden; i++) {
			if (!useCTRNN) {
				double bias = chromosome.getGene(genePos++);
				double a = chromosome.getGene(genePos++) + 2.0;
				if (!simulation) {
					Neuron hiddenNeuron = new SigmoidNeuron(a, bias);
					hiddenNeurons.add(hiddenNeuron);
					net.addHiddenNeuron(hiddenNeuron);
				}
			} else {
				double bias = chromosome.getGene(genePos++);
				double learn_rate = 1.0/((chromosome.getGene(genePos++)+2.0)/4.0);
				if (!simulation) {
					Neuron hiddenNeuron = new CTRNNNeuron(0,learn_rate, bias);
					hiddenNeurons.add(hiddenNeuron);
					net.addHiddenNeuron(hiddenNeuron);
				}

			}
		}

		List<Neuron> outputNeurons = simulation? null:new ArrayList<Neuron>();
		if (!simulation)
		for (int i=0; i < numOutputs; i++) {
			OutputNeuron outputNeuron = new OutputNeuron();
			outputNeurons.add(outputNeuron);
			net.addOutputNeuron(outputNeuron);
		}

		for (int i=0;i<numInputs;i++) {
			for (int j=0;j< numHidden;j++) {
				double weight = chromosome.getGene(genePos++);
				if (!simulation)
				inputNeurons.get(i).addConnection(new CellConnection(weight, hiddenNeurons.get(j)));
			}
		}
		
		if (isRecurrent) {
			for (int i=0;i<numHidden;i++) {
				for (int j=0;j< (numHidden);j++) {	
					double weight = chromosome.getGene(genePos++);
					if (!simulation)
					hiddenNeurons.get(i).addConnection(new CellConnection(weight, (hiddenNeurons.get(j))));
				}
			}			
		}
		
		for (int i=0;i<numHidden;i++) {
			for (int j=0;j< numOutputs;j++) {
				double weight = chromosome.getGene(genePos++);
				if (!simulation)
				hiddenNeurons.get(i).addConnection(new CellConnection(weight, (outputNeurons.get(j))));
			}
		}
		return new Pair<NeuralNetwork, Integer>(net,genePos);
	}
	

	
	
	public static IPlayerTranscriber getPlayerTranscriber(final boolean isRecurrent, final boolean useCTRNN) {
		return new IPlayerTranscriber() {

			public Genome getSampleChromosome() {
				return CompositeSimpleRacer.getSampleChromosome(isRecurrent);
			}

			public Controller getPlayer(Genome genome) {
				return new RacerAdapter(new CompositeSimpleRacer(genome, isRecurrent,useCTRNN));
			}
		};
	}
	
	
	public static Genome getSampleChromosome(boolean isRecurrent) {
		Pair<NeuralNetwork, Integer> networkPair = createNetwork(new Genome(1,0,0), 0, isRecurrent, false, numInputs_SC(), numHidden_SC(), numOutputs_SC(), true);
		networkPair = createNetwork(new Genome(1,0,0), networkPair.second(), isRecurrent, false, numInputs_DC(), numHidden_DC(), numOutputs_DC(),true);
		networkPair = createNetwork(new Genome(1,0,0), networkPair.second(), isRecurrent, false, numInputs_TC(), numHidden_TC(), numOutputs_TC(),true);
		int genomeSize = networkPair.second(); 
		Genome genome = new Genome(genomeSize,-2,2);
		//		System.out.println("gene size:"+genomeSize);
		return genome;
	}
	


	public void reset() {
		speedController.reset();
		directionController.reset();
		targetChooser.reset();
	}
	




	public double[] getAction(GeneticSensorModel model) {
		boolean followNext;
		double distRatio = model.getDistanceToNext()/model.getOtherDistanceToNext();
		distRatio = Math.max(-5,Math.min(5,distRatio));
		double mySpeed = model.getSpeed();
		double otherSpeed = model.getOtherSpeed();
		double speedRatio = mySpeed/otherSpeed;
		speedRatio = Math.max(-5,Math.min(5,speedRatio));
		if (Double.isNaN(speedRatio)) {
			speedRatio = 5;
		}
		if (model.isOthersPresent()) {
//			double[] tcInput = { 
//					model.getWhoReachesNextFirstEstimate(),
//					model.getSpeedToNext(), 
//					model.getOtherSpeedToNext(),
//					model.getDistanceToNext(), 
//					model.getOtherDistanceToNext(), 
//					speedRatio,
//					distRatio};
//					
//			
//		
//			double tcActivation = targetChooser.activate(tcInput, 3)[0];
//			
////			System.out.println(tcActivation);
//			followNext = tcActivation < 0;
//			if (Math.random()<0.00001) {
//				
//				System.out.println("in:"+Genome.arrayToString(tcInput));
//				System.out.println("out:"+tcActivation);
//				System.out.println("my:"+mySpeed+" other:"+otherSpeed);
//			}
			followNext = model.getWhoReachesNextFirstEstimate() == 0;
		} else {
			followNext = true;
		}
		
		
		double angleToTarget = followNext ? model.getAngleToNext() : model.getAngleToNextNext();
		double distanceToTarget = followNext ? model.getDistanceToNext()*10 : model.getDistanceToNextNext()*10;
		double angleToCenter = followNext ? model.getAngleToNextNext(): model.getAngleToCenter();
		double distanceToCenter = followNext ? model.getDistanceToNextNext()*10: model.getDistanceToCenter()*10;
		double[] dcInput = {
								angleToTarget,
								-angleToTarget,
								distanceToTarget,
								mySpeed,
								angleToCenter,
								-angleToCenter,
								distanceToCenter,
//								model.getAngleToOther(),
//								model.getDistanceToOther(),
							};
		double direction = directionController.activate(dcInput, 2)[0];
		
		double[] scInput = {
				angleToTarget,
				Math.abs(angleToTarget),
				distanceToTarget,
				mySpeed,
				angleToCenter,
				Math.abs(angleToCenter),
				distanceToCenter,
//				model.getAngleToOther(),
//				model.getDistanceToOther(),
				};
		double speed = speedController.activate(scInput, 2)[0];
		double[] control = {speed,direction};
		return control;
	}

	public NeuralNetwork[] getNetworks() {
		NeuralNetwork[] networks = {targetChooser, speedController, directionController};
		return networks;
	}

	public Double[] getNetworkUsage() {
		// TODO Auto-generated method stub
		return null;
	}
	
	public static void main(String[] args) throws IOException {
        ObjectWriter writer = new SimpleWriter();
        RacerAdapter racer = new RacerAdapter(new CompositeSimpleRacer());
        Element el = writer.write( racer);
        XMLOutputter out = new XMLOutputter();
        out.output( el, new FileOutputStream("/tmp/racer.xml") );
        System.out.println("");

		
		CompetitionScore cs = new CompetitionScore();
	
		KeyboardControllerOne player1 = new KeyboardControllerOne();
		CompositeSimpleRacer player2 = new CompositeSimpleRacer();
		Evaluator.evaluateDuo(player1, new RacerAdapter(player2), new BasicTrack(), true);
	}
}
