package tomracer;

import simplerace.Car;
import simplerace.Constants;
import simplerace.Controller;
import simplerace.HeuristicController;
import simplerace.SensorModel;
import simplerace.Utils;
import simplerace.Vector2d;
import simplerace.Vehicle;

public class RacerAdapter implements Controller {

	private IGeneticRacer racer;
	private double continuousDirection;
	private double continuousSpeed;

	public RacerAdapter(IGeneticRacer racer) {
		this.racer = racer;
	}
	
	public static double getOtherSpeed ( Vector2d velocity, double orientation ) {
        double velocityMagnitude = velocity.mag ();
        double directionOfMovement = Math.atan2 (velocity.y, velocity.x);
        double directionDifference = directionOfMovement - orientation;
        return Math.cos (directionDifference) * velocityMagnitude;
    }

	static int width = 400;
	static int height = 400;
	static int diagonal = (int) Math.sqrt (width * width + height * height);
    public static double getDistanceToPoint( Vector2d waypoint, Vector2d pos ) {
        return pos.dist(waypoint) / diagonal;
   }
    
    public static double getAngleToPoint (Vector2d nextWp, Vector2d position, double orientation) {
        double xDiff = nextWp.x - position.x;
        double yDiff = nextWp.y - position.y;
        double angle = Math.atan2 (yDiff, xDiff);
        angle = orientation - angle;
        return Utils.correctAngle (angle);
    }

    public static double getSpeedTowards(Vector2d vehiclePos, Vector2d vehicleVelocity, Vector2d point) {
		Vector2d otherPos = vehiclePos;
		Vector2d nextWPPos = point.copy();
		nextWPPos.subtract(otherPos);
		nextWPPos.setMag(1);
		double otherTowardWaypointSpeed =  nextWPPos.x * vehicleVelocity.x + nextWPPos.y * vehicleVelocity.y;
		return otherTowardWaypointSpeed;
    }
    

    public static int whoReachesFirst(Vector2d v1Pos, Vector2d v2Pos, Vector2d v1Vel, Vector2d v2Vel, Vector2d point){
    	double v1Speed = getSpeedTowards(v1Pos, v1Vel, point);
    	double v2Speed = getSpeedTowards(v2Pos, v2Vel, point);
    	double v1Dist = v1Pos.dist(point);
    	double v2Dist = v2Pos.dist(point);
    	if (v1Speed >0 && v2Speed >0) {
    		return v1Dist/v1Speed<v2Dist/v2Speed ? 0:1;
    	}
    	if (v1Speed < 0) {
    		if (v2Speed < 0) {
    			return v1Dist<v2Dist ? 0:1;
    		} else {
    			return 1;
    		}
    	} else {
    		if (v2Speed<0) {
    			return 0;
    		}
    	}
    	return v1Dist<v2Dist?0:1;
    }

	public int control(SensorModel inputs) {
		double orientation = inputs.getOrientation();
		
		Vector2d midPoint = new Vector2d(200,200);
		Vector2d ownPosition = inputs.getPosition();
		Vector2d nextWaypointPosition = inputs.getNextWaypointPosition();
		double angleToNext = getAngleToPoint(nextWaypointPosition, ownPosition, orientation);
		double angleToNextNext = getAngleToPoint(inputs.getNextNextWaypointPosition(), ownPosition, orientation);
		double angleToOther = getAngleToPoint(inputs.getOtherVehiclePosition(), ownPosition, orientation);
		double angleToCenter = getAngleToPoint(midPoint, ownPosition, orientation);
		double distanceToNext = inputs.getDistanceToNextWaypoint();
		double distanceToNextNext = inputs.getDistanceToNextNextWaypoint();
		double distanceToOther = inputs.getDistanceToOtherVehicle();
		double distanceToCenter = getDistanceToPoint(midPoint, ownPosition); 
		double otherDistanceToNext = getDistanceToPoint(nextWaypointPosition, inputs.getOtherVehiclePosition());
		double speed = getOtherSpeed(inputs.getVelocity(),inputs.getOrientation());//inputs.getSpeed();
//		if (Math.random()<0.0001) {
//			System.out.println("mySpppeeeed:"+speed);
//		}
		double speedToNext = getSpeedTowards(inputs.getPosition(), inputs.getNextWaypointPosition(), inputs.getNextWaypointPosition());
		double otherSpeed = getOtherSpeed(inputs.getOtherVehicleVelocity(), inputs.getOtherVehicleOrientation());
		double otherSpeedToNext = getSpeedTowards(inputs.getOtherVehiclePosition(), inputs.getOtherVehicleVelocity(), inputs.getNextWaypointPosition());
		boolean othersPresent = inputs.otherVehicleIsPresent();

		int whoReachesNextFirstEstimate = !othersPresent ? -1 :whoReachesFirst(inputs.getPosition(), inputs.getOtherVehiclePosition(), inputs.getVelocity(), inputs.getOtherVehicleVelocity(), inputs.getNextWaypointPosition());
//whoReachesFirstSimulation(inputs.getPosition(), inputs.getOtherVehiclePosition(), inputs.getVelocity(), inputs.getOtherVehicleVelocity(), inputs.getOrientation(), inputs.getOtherVehicleOrientation(), inputs.getNextWaypointPosition());
		GeneticSensorModel model = new GeneticSensorModel(angleToNext, angleToNextNext, angleToOther, angleToCenter, distanceToNext, distanceToNextNext, distanceToOther, distanceToCenter, otherDistanceToNext, speed, speedToNext, otherSpeed, otherSpeedToNext, othersPresent, whoReachesNextFirstEstimate);
		
		double[] action = racer.getAction(model);
		
		

		
		continuousSpeed += action[0];
		int outSpeed;
		if (continuousSpeed>1) {
			outSpeed =2;
			continuousSpeed=0;
		}
		else
		if (continuousSpeed<-1) {
			outSpeed =0;
			continuousSpeed = 0;
		}
		else {
			outSpeed = 1;
		}		

		continuousDirection += action[1];
		int outDirection;
		if (continuousDirection>1) {
			outDirection =2;
			continuousDirection=0;
		}
		else
		if (continuousDirection<-1) {
			outDirection =0;
			continuousDirection = 0;
		}
		else {
			outDirection = 1;
		}
		

		return outSpeed*3+outDirection;
	}

	public void reset() {
		racer.reset();
		continuousSpeed=0;
		continuousDirection=0;
	}
	public IGeneticRacer getRacer() {
		return racer;
	}
	

}


